import multiprocessing.pool
import os
import subprocess as sp
from datetime import datetime

import pandas as pd
import pyrodigal
import pyrodigal_gv
import pyrodigal_rv
from BCBio import GFF
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from external_tools import ExternalTool
from loguru import logger
from util import parse_attributes, remove_directory


def run_pyrodigal_gv(filepath_in, out_dir, threads):
    """
    Gets CDS using pyrodigal_gv
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param logger logger
    :param meta Boolean - metagenomic mode flag
    :param coding_table coding table for prodigal (default 11)
    :return:
    """

    # true
    orf_finder = pyrodigal_gv.ViralGeneFinder(meta=True)

    def _find_genes(record):
        genes = orf_finder.find_genes(str(record.seq))
        return (record.id, genes)

    with multiprocessing.pool.ThreadPool(threads) as pool:
        with open(os.path.join(out_dir, "prodigal-gv_out.gff"), "w") as gff:
            with open(os.path.join(out_dir, "prodigal-gv_out_tmp.fasta"), "w") as dst:
                with open(
                    os.path.join(out_dir, "prodigal-gv_out_aas_tmp.fasta"), "w"
                ) as aa_fasta:
                    records = SeqIO.parse(filepath_in, "fasta")
                    for record_id, genes in pool.imap(_find_genes, records):
                        genes.write_gff(
                            gff, sequence_id=record_id, include_translation_table=True
                        )
                        genes.write_genes(dst, sequence_id=record_id)
                        # need to write the translation
                        genes.write_translations(aa_fasta, sequence_id=record_id)


def run_pyrodigal_rv(filepath_in, out_dir, threads):
    """
    Gets CDS using pyrodigal_gv
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param logger logger
    :param meta Boolean - metagenomic mode flag
    :param coding_table coding table for prodigal (default 11)
    :return:
    """

    # true
    orf_finder = pyrodigal_rv.ViralGeneFinder(meta=True)

    def _find_genes(record):
        genes = orf_finder.find_genes(str(record.seq))
        return (record.id, genes)

    with multiprocessing.pool.ThreadPool(threads) as pool:
        with open(os.path.join(out_dir, "pyrodigal-rv_out.gff"), "w") as gff:
            with open(os.path.join(out_dir, "pyrodigal-rv_out_tmp.fasta"), "w") as dst:
                with open(
                    os.path.join(out_dir, "pyrodigal-rv_out_aas_tmp.fasta"), "w"
                ) as aa_fasta:
                    records = SeqIO.parse(filepath_in, "fasta")
                    for record_id, genes in pool.imap(_find_genes, records):
                        genes.write_gff(
                            gff, sequence_id=record_id, include_translation_table=True
                        )
                        genes.write_genes(dst, sequence_id=record_id)
                        # need to write the translation
                        genes.write_translations(aa_fasta, sequence_id=record_id)


##### phanotate meta mode ########


def batch_iterator(iterator, batch_size):
    """Returns lists of length batch_size.
    https://biopython.org/wiki/Split_large_file

    :param iterator: iterator for enumerating over
    :param batch_size: number of fasta records in each file
    :return:
    """
    batch = []
    for entry in iterator:
        batch.append(entry)
        if len(batch) == batch_size:
            yield batch
            batch = []


def split_input_fasta(filepath_in, out_dir):
    """Splits the input fasta into separate single fasta files for multithreading with phanotate
    https://biopython.org/wiki/Split_large_file

    :param filepath_in: input multifasta file
    :param out_dir: output director
    :return: num_fastas: int giving the number of fasta records in the multifasta
    """
    # iterate and count fastas
    record_iter = SeqIO.parse(open(filepath_in), "fasta")
    num_fastas = len([1 for line in open(filepath_in) if line.startswith(">")])

    # each fasta gets its own file so batch size of 1
    batch_size = 1

    input_tmp_dir = os.path.join(out_dir, "input_split_tmp")

    for i, batch in enumerate(batch_iterator(record_iter, batch_size)):
        filename = "input_subprocess%i.fasta" % (i + 1)
        with open(os.path.join(input_tmp_dir, filename), "w") as handle:
            SeqIO.write(batch, handle, "fasta")
    return num_fastas


def run_phanotate_fasta_meta(filepath_in, out_dir, threads, num_fastas):
    """
    Runs phanotate to output fastas
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param threads: threads
    :param num_fastas: number of fastas in input multifasta
    :return:
    """

    phanotate_tmp_dir = os.path.join(out_dir, "input_split_tmp")
    commands = []

    for i in range(1, num_fastas + 1):
        in_file = "input_subprocess" + str(i) + ".fasta"
        out_file = "phanotate_out_tmp" + str(i) + ".fasta"
        filepath_in = os.path.join(phanotate_tmp_dir, in_file)
        cmd = (
            "phanotate.py "
            + filepath_in
            + " -o "
            + os.path.join(phanotate_tmp_dir, out_file)
            + " -f fasta"
        )
        commands.append(cmd)

    n = int(threads)  # the number of parallel processes you want

    for j in range(max(int(len(commands) / n) + 1, 1)):
        procs = [
            sp.Popen(i, shell=True)
            for i in commands[j * n : min((j + 1) * n, len(commands))]
        ]
        for p in procs:
            p.wait()


def run_phanotate_txt_meta(filepath_in, out_dir, threads, num_fastas):
    """
    Runs phanotate to output text file
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param threads: threads
    :param num_fastas: number of fastas in input multifasta
    :return:
    """

    phanotate_tmp_dir = os.path.join(out_dir, "input_split_tmp")

    commands = []

    for i in range(1, num_fastas + 1):
        in_file = "input_subprocess" + str(i) + ".fasta"
        out_file = "phanotate_out_tmp" + str(i) + ".txt"
        filepath_in = os.path.join(phanotate_tmp_dir, in_file)
        cmd = (
            "phanotate.py "
            + filepath_in
            + " -o "
            + os.path.join(phanotate_tmp_dir, out_file)
            + " -f tabular"
        )
        commands.append(cmd)

    n = int(threads)  # the number of parallel processes you want
    for j in range(max(int(len(commands) / n) + 1, 1)):
        procs = [
            sp.Popen(i, shell=True)
            for i in commands[j * n : min((j + 1) * n, len(commands))]
        ]
        for p in procs:
            p.wait()


def concat_phanotate_meta(out_dir, num_fastas):
    """
    Concatenates phanotate output for downstream analysis
    :param out_dir: output directory
    :param threads: threads
    :return:
    """

    phanotate_tmp_dir = os.path.join(out_dir, "input_split_tmp")

    tsvs = []
    for i in range(1, int(num_fastas) + 1):
        out_tsv = "phanotate_out_tmp" + str(i) + ".txt"
        tsvs.append(os.path.join(phanotate_tmp_dir, out_tsv))

    with open(os.path.join(out_dir, "phanotate_out.txt"), "w") as outfile:
        for fname in tsvs:
            with open(fname) as infile:
                outfile.write(infile.read())

    fastas = []
    for i in range(1, int(num_fastas) + 1):
        out_fasta = "phanotate_out_tmp" + str(i) + ".fasta"
        fastas.append(os.path.join(phanotate_tmp_dir, out_fasta))

    with open(os.path.join(out_dir, "phanotate_out_tmp.fasta"), "w") as outfile:
        for fname in fastas:
            with open(fname) as infile:
                outfile.write(infile.read())


def run_trnascan_meta(filepath_in, out_dir, threads, num_fastas, trna_scan_model):
    """
    Runs trnascan to output gffs one contig per thread
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param threads: threads
    :param num_fastas: number of fastas in input multifasta
    :return:
    """

    input_tmp_dir = os.path.join(out_dir, "input_split_tmp")
    commands = []

    if trna_scan_model == "general":
        model = "G"
    else:  # bacterial
        model = "B"

    for i in range(1, num_fastas + 1):
        in_file = "input_subprocess" + str(i) + ".fasta"
        out_file = "trnascan_tmp" + str(i) + ".gff"
        sec_struc_file = "trnascan_tmp" + str(i) + ".sec"
        filepath_in = os.path.join(input_tmp_dir, in_file)
        filepath_out = os.path.join(input_tmp_dir, out_file)
        sec_out = os.path.join(input_tmp_dir, sec_struc_file)
        cmd = f"tRNAscan-SE {filepath_in} --thread 1 -{model} -D -Q -j {filepath_out} -f {sec_out}"
        commands.append(cmd)

    n = int(threads)  # the number of parallel processes you want

    for j in range(max(int(len(commands) / n) + 1, 1)):
        procs = [
            sp.Popen(i, shell=True, stderr=sp.PIPE, stdout=sp.DEVNULL)
            for i in commands[j * n : min((j + 1) * n, len(commands))]
        ]
        for p in procs:
            p.wait()


def concat_trnascan_meta(out_dir, num_fastas):
    """
    Concatenates trnascan output for downstream analysis
    :param out_dir: output directory
    :param threads: threads
    :return:
    """

    input_tmp_dir = os.path.join(out_dir, "input_split_tmp")

    gffs = []
    for i in range(1, int(num_fastas) + 1):
        out_gff = "trnascan_tmp" + str(i) + ".gff"
        gffs.append(os.path.join(input_tmp_dir, out_gff))

    with open(os.path.join(out_dir, "trnascan_out.gff"), "w") as outfile:
        for fname in gffs:
            with open(fname) as infile:
                outfile.write(infile.read())

    sec_strucs = []
    for i in range(1, int(num_fastas) + 1):
        sec_struc_file = "trnascan_tmp" + str(i) + ".sec"
        sec_strucs.append(os.path.join(input_tmp_dir, sec_struc_file))

    with open(os.path.join(out_dir, "trnascan_out.sec"), "w") as outfile:
        for fname in sec_strucs:
            with open(fname) as infile:
                outfile.write(infile.read())


##### single contig mode ######


def run_phanotate(filepath_in, out_dir, logdir):
    """
    Runs phanotate
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param logdir logdir
    :return:
    """

    out_fasta = os.path.join(out_dir, "phanotate_out_tmp.fasta")
    out_tab = os.path.join(out_dir, "phanotate_out.txt")

    phan_fast = ExternalTool(
        tool="phanotate.py",
        input=f"{filepath_in}",
        output=f"-o {out_fasta}",
        params=f"-f fasta",
        logdir=logdir,
        outfile="",
    )

    phan_txt = ExternalTool(
        tool="phanotate.py",
        input=f"{filepath_in}",
        output=f"-o {out_tab}",
        params=f"-f tabular",
        logdir=logdir,
        outfile="",
    )

    try:
        ExternalTool.run_tool(phan_fast)
        ExternalTool.run_tool(phan_txt)

    except:
        logger.error("Error with Phanotate\n")


def run_pyrodigal(filepath_in, out_dir, meta, coding_table, threads):
    """
    Gets CDS using pyrodigal
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param logger logger
    :param meta Boolean - metagenomic mode flag
    :param coding_table coding table for prodigal (default 11)
    :param threads: threads
    :return:
    """

    prodigal_metamode = False
    if meta == True:
        prodigal_metamode = True
        logger.info("Prodigal Meta Mode Enabled")

    #######################################################
    # if under 20000, pyrodigal will only work in meta mode
    # https://github.com/hyattpd/prodigal/wiki/Advice-by-Input-Type#plasmids-phages-viruses-and-other-short-sequences
    # https://github.com/hyattpd/Prodigal/issues/51
    # so make sure of this
    #######################################################

    # get total length of input
    total_length = 0

    with open(filepath_in, "r") as handle:
        for record in SeqIO.parse(handle, "fasta"):
            total_length += len(record.seq)

    # if the length is 20000 or under, use meta mode by default
    # reduce from 100000 #https://github.com/gbouras13/pharokka/issues/409
    if total_length < 100001:
        orf_finder = pyrodigal.GeneFinder(meta=True)
        prodigal_metamode = True

    # otherwise train it
    # recommend pyrodigal-gv anyway
    else:
        # for training if you want different coding table
        seqs = [bytes(record.seq) for record in SeqIO.parse(filepath_in, "fasta")]
        record = SeqIO.parse(filepath_in, "fasta")
        orf_finder = pyrodigal.GeneFinder(meta=prodigal_metamode)

        # make coding table possible if false
        if prodigal_metamode == False:
            orf_finder.train(*seqs, translation_table=int(coding_table))

    # define for the multithreadpool
    def _find_genes(record):
        genes = orf_finder.find_genes(str(record.seq))
        return (record.id, genes)

    if prodigal_metamode:
        transl_table = None 
    else:
        transl_table = "11" 


    with multiprocessing.pool.ThreadPool(threads) as pool:
        with open(os.path.join(out_dir, "prodigal_out.gff"), "w") as gff:
            with open(os.path.join(out_dir, "prodigal_out_tmp.fasta"), "w") as dst:
                with open(
                    os.path.join(out_dir, "prodigal_out_aas_tmp.fasta"), "w"
                ) as aa_fasta:
                    records = SeqIO.parse(filepath_in, "fasta")
                    for record_id, genes in pool.imap(_find_genes, records):
                        genes.write_gff(gff, sequence_id=record_id)
                        genes.write_genes(dst, sequence_id=record_id)
                        # need to write the translation
                        genes.write_translations(aa_fasta, sequence_id=record_id)

                        # get transl_table once
                        if transl_table is None:
                            for gene in genes:
                                transl_table = str(gene.translation_table)
                                break
    return str(transl_table)

def tidy_phanotate_output(out_dir):
    """
    Tidies phanotate output
    :param out_dir: output directory
    :return: phan_df pandas dataframe
    """
    phan_file = os.path.join(out_dir, "phanotate_out.txt")
    col_list = ["start", "stop", "strand", "contig", "score"]
    dtype_dict = {
        "start": int,
        "stop": int,
        "strand": str,
        "contig": str,
        "score": float,
    }

    phan_df = pd.read_csv(
        phan_file,
        delimiter="\t",
        index_col=False,
        names=col_list,
        skiprows=2,  # to skip the headers
        dtype=dtype_dict,
        comment="#",
    )
    # get rid of the headers and reset the index
    phan_df = phan_df[phan_df["start"] != "#id:"]
    phan_df = phan_df[phan_df["start"] != "#START"].reset_index(drop=True)
    phan_df["gene"] = (
        phan_df["contig"].astype(str)
        + phan_df.index.astype(str)
        + " "
        + phan_df["start"].astype(str)
        + "_"
        + phan_df["stop"].astype(str)
    )
    phan_df.to_csv(
        os.path.join(out_dir, "cleaned_phanotate.tsv"), sep="\t", index=False
    )
    return phan_df


def tidy_prodigal_output(out_dir, gene_predictor):
    """
    Tidies prodigal output
    :param out_dir: output directory
    :param gene_predictor: gene predictor
    :return: prod_filt_df pandas dataframe
    """

    prefix = gene_predictor

    prod_file = os.path.join(out_dir, f"{prefix}_out.gff")
    col_list = [
        "contig",
        "prod",
        "orf",
        "start",
        "stop",
        "score",
        "strand",
        "frame",
        "description",
    ]
    dtype_dict = {
        "contig": str,
        "prod": str,
        "orf": str,
        "start": int,
        "stop": int,
        "score": float,
        "strand": str,
        "frame": str,
        "description": str,
    }

    prod_df = pd.read_csv(
        prod_file,
        delimiter="\t",
        index_col=False,
        names=col_list,
        dtype=dtype_dict,
        comment="#",  # to skip the headers
    )

    # meta mode brings in some Nas so remove them
    # need to reset index!!!! and drop, or else will cause rubbish results for metagenomics
    prod_df = prod_df.dropna().reset_index(drop=True)

    # keep description column and parse it to access "partial" info
    prod_filt_df = prod_df[
        ["start", "stop", "strand", "contig", "score", "description"]
    ]

    parsed = prod_filt_df["description"].apply(parse_attributes)

    # Convert the list of dicts into a DataFrame
    attributes_df = pd.DataFrame(parsed.tolist())

    # Drop score column to avoid double score columns
    attributes_df = attributes_df.drop(columns=["score"])

    # Concatenate with the original DataFrame (if you want to keep other columns)
    prod_filt_df = pd.concat(
        [prod_filt_df.reset_index(drop=True), attributes_df], axis=1
    )

    prod_filt_df = prod_filt_df[
        ["start", "stop", "strand", "contig", "score", "partial"]
    ]

    # convert start stop to string
    prod_filt_df["start"] = prod_filt_df["start"].astype(int)
    prod_filt_df["stop"] = prod_filt_df["stop"].astype(int)

    # rearrange start and stop so that for negative strand, the stop is before start (like phanotate_out)
    cols = ["start", "stop"]
    # indices where start is greater than stop
    ixs = prod_filt_df["strand"] == "-"
    # Where ixs is True, values are swapped
    prod_filt_df.loc[ixs, cols] = (
        prod_filt_df.loc[ixs, cols].reindex(columns=cols[::-1]).values
    )

    prod_filt_df["gene"] = (
        prod_filt_df["contig"].astype(str)
        + prod_filt_df.index.astype(str)
        + " "
        + prod_filt_df["start"].astype(str)
        + "_"
        + prod_filt_df["stop"].astype(str)
    )
    prod_filt_df.to_csv(
        os.path.join(out_dir, f"cleaned_{prefix}.tsv"), sep="\t", index=False
    )
    return prod_filt_df


def tidy_genbank_output(out_dir, genbank_file, coding_table):
    """
    Tidies --genbank input output
    :param out_dir: output directory
    :return: phan_df pandas dataframe
    """

    # Create lists to store CDS information
    starts = []
    stops = []
    strands = []
    contigs = []

    # output nuc
    fasta_nucl_tmp = "genbank_out_tmp.fasta"
    # output aa file
    fasta_output_aas_tmp = "genbank_aas_tmp.fasta"

    # make dataframe
    for record in SeqIO.parse(genbank_file, "genbank"):
        for feature in record.features:
            if feature.type == "CDS":
                strand = feature.location.strand
                if strand == 1:  # pos
                    strand = "+"
                    start = (
                        feature.location.start + 1
                    )  # this needs to added by 1 - the parser is 0 indexed issue #353
                    stop = feature.location.end
                else:  # neg
                    strand = "-"
                    start = feature.location.end
                    stop = (
                        feature.location.start + 1
                    )  # this needs to added by 1 - the parser is 0 indexed issue #353

                contig = record.id
                starts.append(start)
                stops.append(stop)
                strands.append(strand)
                contigs.append(contig)

    # Create a pandas DataFrame
    data = {"start": starts, "stop": stops, "strand": strands, "contig": contigs}

    gen_df = pd.DataFrame(data)
    # add fake score
    gen_df["score"] = "No_score"

    # get the gene
    gen_df["gene"] = (
        gen_df["contig"].astype(str)
        + gen_df.index.astype(str)
        + " "
        + gen_df["start"].astype(str)
        + "_"
        + gen_df["stop"].astype(str)
    )

    gen_df.to_csv(os.path.join(out_dir, "cleaned_genbank.tsv"), sep="\t", index=False)

    # get temporary AA output
    with open(os.path.join(out_dir, fasta_nucl_tmp), "w") as nt_fa:
        with open(os.path.join(out_dir, fasta_output_aas_tmp), "w") as aa_fa:
            i = 0
            for record in SeqIO.parse(genbank_file, "genbank"):
                for feature in record.features:
                    if feature.type == "CDS":
                        header = str(gen_df["contig"].iloc[i]) + str(i)
                        description = (
                            str(gen_df["start"].iloc[i])
                            + "_"
                            + str(gen_df["stop"].iloc[i])
                        )

                        # Extract NT and CDS sequence
                        cds_nt_seq = feature.location.extract(record.seq)
                        cds_amino_acid_seq = cds_nt_seq.translate(
                            to_stop=True, table=coding_table
                        )

                        # save to file
                        nt_record = aa_record = SeqRecord(
                            cds_nt_seq,
                            id=header,
                            description=description,
                        )

                        aa_record = aa_record = SeqRecord(
                            cds_amino_acid_seq,
                            id=header,
                            description=description,
                        )

                        SeqIO.write(aa_record, aa_fa, "fasta")
                        SeqIO.write(nt_record, nt_fa, "fasta")
                        i += 1

    return gen_df


def translate_fastas(out_dir, gene_predictor, coding_table, genbank_file):
    """
    Translates input CDSs to amino acids. For genbank, it will just output the CDS coordinates in the file
    :param out_dir: output directory
    :param gene_predictor: phanotate or prodigal
    :param genbank_file: genbank file if gene_predictor is genbank
    :return:
    """
    if gene_predictor == "phanotate":
        clean_df = tidy_phanotate_output(out_dir)
    elif gene_predictor == "prodigal" or gene_predictor == "prodigal-gv" or gene_predictor == "pyrodigal-rv" :
        clean_df = tidy_prodigal_output(out_dir, gene_predictor)  
    elif gene_predictor == "genbank":
        clean_df = tidy_genbank_output(out_dir, genbank_file, coding_table)

    fasta_output_aas_tmp = gene_predictor + "_aas_tmp.fasta"

    if gene_predictor == "phanotate":
        # read the nucl fasta
        fasta_input_tmp = gene_predictor + "_out_tmp.fasta"
        # translate for temporary AA output
        with open(os.path.join(out_dir, fasta_output_aas_tmp), "w") as aa_fa:
            i = 0
            for dna_record in SeqIO.parse(
                os.path.join(out_dir, fasta_input_tmp), "fasta"
            ):
                dna_header = str(clean_df["contig"].iloc[i]) + str(i)
                dna_description = (
                    str(clean_df["start"].iloc[i]) + "_" + str(clean_df["stop"].iloc[i])
                )
                aa_record = SeqRecord(
                    dna_record.seq.translate(to_stop=True, table=coding_table),
                    id=dna_header,
                    description=dna_description,
                )
                SeqIO.write(aa_record, aa_fa, "fasta")
                i += 1
    elif gene_predictor == "prodigal-gv" or gene_predictor == "prodigal" or gene_predictor == "pyrodigal-rv":
        # read in the AA file instead and parse that to clean the header
        fasta_input_tmp = gene_predictor + "_out_aas_tmp.fasta"
        with open(os.path.join(out_dir, fasta_output_aas_tmp), "w") as aa_fa:
            i = 0
            for dna_record in SeqIO.parse(
                os.path.join(out_dir, fasta_input_tmp), "fasta"
            ):
                dna_header = str(clean_df["contig"].iloc[i]) + str(i)
                dna_description = (
                    str(clean_df["start"].iloc[i]) + "_" + str(clean_df["stop"].iloc[i])
                )
                aa_record = SeqRecord(
                    dna_record.seq,
                    id=dna_header,
                    description=dna_description,
                )
                SeqIO.write(aa_record, aa_fa, "fasta")
                i += 1
    # for genbank do nothing


def run_trna_scan(filepath_in, threads, out_dir, logdir, trna_scan_model):
    """
    Runs trna scan
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param logger logger
    :return:
    """

    out_gff = os.path.join(out_dir, "trnascan_out.gff")
    out_sec = os.path.join(out_dir, "trnascan_out.sec")

    if trna_scan_model == "general":
        model = "G"
    else:  # bacterial
        model = "B"

    trna = ExternalTool(
        tool="tRNAscan-SE",
        input=f"{filepath_in}",
        output=f"{out_gff}",
        params=f"--thread {threads} -{model} -f {out_sec} -Q -j",
        logdir=logdir,
        outfile="",
    )

    try:
        ExternalTool.run_tool(trna)
    except:
        logger.error("Error with tRNAscan-SE")
        return 0


def run_mmseqs(db_dir, out_dir, threads, logdir, gene_predictor, evalue, reverse_mmseqs2, sensitivity, db_name):
    """
    Runs mmseqs2 on phrogs
    :param db_dir: database path
    :param out_dir: output directory
    :param logger: logger
    :params threads: threads
    :param gene_predictor: phanotate or prodigal
    :param evalue: evalue for mmseqs2
    :param reverse_mmseqs2: reverse MMseqs2 database to be target not query
    :param sensitivity: MMseqs2 sensitivity
    :param db_name: str one of 'PHROG', 'VFDB' or 'CARD'
    :return:
    """

    logger.info(f"Running MMseqs2 on {db_name} Database.")

    # declare files
    amino_acid_fasta = gene_predictor + "_aas_tmp.fasta"

    # define the outputs
    if db_name == "PHROG":
        mmseqs_dir = os.path.join(out_dir, "mmseqs/")
        target_db_dir = os.path.join(out_dir, "target_dir/")
        tmp_dir = os.path.join(out_dir, "tmp_dir/")
        profile_db = os.path.join(db_dir, "phrogs_profile_db")
        mmseqs_result_tsv = os.path.join(out_dir, "mmseqs_results.tsv")
    elif db_name == "VFDB":
        mmseqs_dir = os.path.join(out_dir, "VFDB/")
        target_db_dir = os.path.join(out_dir, "VFDB_target_dir/")
        tmp_dir = os.path.join(out_dir, "VFDB_dir/")
        profile_db = os.path.join(db_dir, "vfdb")
        mmseqs_result_tsv = os.path.join(out_dir, "vfdb_results.tsv")
    elif db_name == "CARD":
        mmseqs_dir = os.path.join(out_dir, "CARD/")
        target_db_dir = os.path.join(out_dir, "CARD_target_dir/")
        tmp_dir = os.path.join(out_dir, "CARD_dir/")
        profile_db = os.path.join(db_dir, "CARD")
        mmseqs_result_tsv = os.path.join(out_dir, "CARD_results.tsv")

    input_aa_fasta = os.path.join(out_dir, amino_acid_fasta)
    target_seqs = os.path.join(target_db_dir, "target_seqs")

    # make dir for target db
    if os.path.isdir(target_db_dir) == False:
        os.mkdir(target_db_dir)

    # creates db for input
    mmseqs_createdb = ExternalTool(
        tool="mmseqs createdb",
        input=f"",
        output=f"{target_seqs}",
        params=f"{input_aa_fasta}",  # param goes before output and mmseqs2 required order
        logdir=logdir,
        outfile="",
    )

    ExternalTool.run_tool(mmseqs_createdb)

    # runs the mmseqs seacrh

    result_mmseqs = os.path.join(mmseqs_dir, "results_mmseqs")

    

    if db_name == "PHROG":
        if reverse_mmseqs2:
            params_list = f"-e {evalue} {target_seqs} {profile_db} {result_mmseqs}"
        
        else:
            params_list = f"-e {evalue} {profile_db} {target_seqs} {result_mmseqs}" # param goes before output and mmseqs2 required order

        mmseqs_search = ExternalTool(
            tool="mmseqs search",
            input=f"",
            output=f"{tmp_dir} -s {sensitivity} --threads {threads}",
            params=params_list,  
            logdir=logdir,
            outfile="",
        )
    else:  # if it is vfdb or card search with cutoffs instead of evalue

        if reverse_mmseqs2:
            params_list = f"--min-seq-id 0.8 -c 0.4 {target_seqs} {profile_db} {result_mmseqs}"
        
        else:
            params_list = f"--min-seq-id 0.8 -c 0.4 {profile_db} {target_seqs} {result_mmseqs}" # param goes before output and mmseqs2 required order


        mmseqs_search = ExternalTool(
            tool="mmseqs search",
            input=f"",
            output=f"{tmp_dir} -s {sensitivity} --threads {threads}",
            params=params_list,  # param goes before output and mmseqs2 required order
            logdir=logdir,
            outfile="",
        )

    ExternalTool.run_tool(mmseqs_search)


    if reverse_mmseqs2:
        params_list = f"{target_seqs} {profile_db} {result_mmseqs}"
    
    else:
        params_list = f"{profile_db} {target_seqs} {result_mmseqs}" # param goes before output and mmseqs2 required order


    # creates the output tsv
    mmseqs_createtsv = ExternalTool(
        tool="mmseqs createtsv",
        input=f"",
        output=f"{mmseqs_result_tsv} --full-header --threads {threads}",
        params=params_list,  # param goes before output and mmseqs2 required order
        logdir=logdir,
        outfile="",
    )

    ExternalTool.run_tool(mmseqs_createtsv)

    # remove the target dir when finished
    remove_directory(target_db_dir)


def convert_gff_to_gbk(filepath_in, input_dir, out_dir, prefix, prot_seq_df):
    """
    Converts the gff to genbank
    :param filepath_in: input fasta file
    :param input_dir: input directory of the gff. same as output_dir for the overall gff in normal mode, differeny for meta mode
    :param out_dir: output directory of the gbk
    :param prefix: prefix
    :param prefix: prot_seq_df from pharok object with gene name + protein sequence for all genes (from create_gff()).
    :return:
    """

    gff_file = os.path.join(input_dir, f"{prefix}.gff")
    gbk_file = os.path.join(out_dir, f"{prefix}.gbk")

    prot_seq_df["contig"] = prot_seq_df["contig"].astype(str)

    with open(gbk_file, "wt") as gbk_handler:
        fasta_handler = SeqIO.to_dict(SeqIO.parse(filepath_in, "fasta"))

        for record in GFF.parse(gff_file, fasta_handler):
            # sequence in each contig (record)
            record.id = str(record.id)
            sequence_length = len(record.seq)
            subset_seqs_df = prot_seq_df.loc[prot_seq_df["contig"] == record.id]
            # get all the seqs in the contigs - and drop the index to reset for 0 indexed loop
            subset_seqs = subset_seqs_df["sequence"].reset_index(drop=True)
            # start the loop
            i = 0

            # instantiate record
            record.annotations["molecule_type"] = "DNA"
            record.annotations["date"] = datetime.today()
            record.annotations["topology"] = "linear"
            record.annotations["data_file_division"] = (
                "PHG"  # https://github.com/RyanCook94/inphared/issues/22
            )

            # add features to the record
            for feature in record.features:
                # Move 'source' qualifier to 'inference', then remove 'source'
                if "source" in feature.qualifiers:
                    feature.qualifiers["inference"] = feature.qualifiers["source"]
                    del feature.qualifiers["source"]

                # If 'anticodon' is present, join multiple parts into one comma-separated string
                if "anticodon" in feature.qualifiers:
                    # Join multiple parts into one comma-separated string
                    if isinstance(feature.qualifiers["anticodon"], list):
                        feature.qualifiers["anticodon"] = [
                            ",".join(feature.qualifiers["anticodon"])
                        ]
                # add translation only if CDS
                if feature.type == "CDS":
                    # fix frame for partial CDS
                    if "phase" in feature.qualifiers:
                        del feature.qualifiers["phase"]

                    if "partial" in feature.qualifiers:
                        if (
                            feature.qualifiers["partial"] == ["10"]
                            and feature.location.strand == 1
                            and feature.location.start
                            != 0  # codon_start of 1 is default, so not necessary to add
                        ):
                            feature.qualifiers["codon_start"] = (
                                feature.location.start + 1
                            )
                        elif (
                            feature.qualifiers["partial"] == ["01"]
                            and feature.location.strand == -1
                        ):
                            feature.qualifiers["codon_start"] = (
                                sequence_length - feature.location.end + 1
                            )

                    feature.qualifiers.update(
                        {"translation": subset_seqs[i]}  # from the aa seq
                    )
                    i += 1

            SeqIO.write(record, gbk_handler, "genbank")


def run_minced(filepath_in, out_dir, prefix, minced_args, logdir):
    """
    Runs MinCED
    :param filepath_in: input fasta file
    :param out_dir: output directory
    :param logger: logger
    :param minced_args: str with extra arguments to pass to MINced
    :params prefix: prefix
    :return:
    """

    logger.info("Running MinCED.")

    output_spacers = os.path.join(out_dir, prefix + "_minced_spacers.txt")
    output_gff = os.path.join(out_dir, prefix + "_minced.gff")

    if minced_args != "":
        minced_args = f"-{minced_args}"

    minced_fast = ExternalTool(
        tool="minced",
        input=f"",
        output=f" {output_spacers} {output_gff}",
        params=f" {minced_args} {filepath_in}",  # need strange order for minced params go first
        logdir=logdir,
        outfile="",
    )

    try:
        ExternalTool.run_tool(minced_fast)
    except:
        logger.error("Error with MinCED\n")


def run_aragorn(filepath_in, out_dir, prefix, logdir):
    """
    Runs run_aragorn
    :param filepath_in: input fasta file
    :param out_dir: output directory
    :param logdir: logdir
    :params prefix: prefix
    :return:
    """
    logger.info("Running Aragorn.")

    aragorn_out_file = os.path.join(out_dir, prefix + "_aragorn.txt")
    aragorn = ExternalTool(
        tool="aragorn",
        input=f"{filepath_in}",
        output=f"-o {aragorn_out_file}",
        params=f"-l -gcbact -w -m",
        logdir=logdir,
        outfile="",
    )

    try:
        ExternalTool.run_tool(aragorn)
    except:
        logger.error("Error with Aragorn\n")


def reorient_terminase(filepath_in, out_dir, prefix, terminase_strand, terminase_start):
    """
    re-orients phage to begin with large terminase subunit
    :param filepath_in input genome fasta
    :param out_dir: output directory path
    :param prefix: prefix for pharokka
    :param terminase_strand: strandedness of the terminase large subunit. Is either 'pos' or 'neg'
    :param terminase_start: start coordinate of terminase large subunit.
    :logger: logger
    """

    logger.info(
        "Input checked. \nReorienting input genome to begin with terminase large subunit."
    )

    # read in the fasta
    record = SeqIO.read(filepath_in, "fasta")

    # get length of the fasta
    length = len(record.seq)

    if int(terminase_start) > length or int(terminase_start) < 1:
        logger.error(
            "Error: terminase large subunit start coordinate specified is not within the provided genome length. Please check your input. \n"
        )

    # positive

    # reorient to start at the terminase
    # pos
    if terminase_strand == "pos":
        start = record.seq[(int(terminase_start) - 1) : length]
        end = record.seq[0 : int(terminase_start) - 1]
        total = start + end

    # neg
    if terminase_strand == "neg":
        record.seq = record.seq.reverse_complement()
        start = record.seq[(length - int(terminase_start)) : length]
        end = record.seq[0 : (length - int(terminase_start))]
        total = start + end

    # set sequence
    record.seq = total

    out_fasta = os.path.join(out_dir, prefix + "_genome_terminase_reoriented.fasta")

    SeqIO.write(record, out_fasta, "fasta")


def run_mash_sketch(filepath_in, out_dir, logdir):
    """
    Runs mash sketch
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param logger logger
    :return:
    """

    mash_sketch_out_file = os.path.join(out_dir, "input_mash_sketch.msh")

    mash_sketch = ExternalTool(
        tool="mash sketch",
        input=f"",
        output=f"-o {mash_sketch_out_file} -i",
        params=f"{filepath_in}",
        logdir=logdir,
        outfile="",
    )

    try:
        ExternalTool.run_tool(mash_sketch)

    except:
        logger.error("Error with mash sketch\n")


def run_mash_dist(out_dir, db_dir, mash_distance, logdir):
    """
    Runs mash
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param mash_distance: mash distance - float
    :param logger logger
    :return:
    """

    mash_sketch = os.path.join(out_dir, "input_mash_sketch.msh")
    phrog_sketch = os.path.join(db_dir, "9Aug2025_genomes.fa.msh")
    mash_tsv = os.path.join(out_dir, "mash_out.tsv")

    mash_dist = ExternalTool(
        tool="mash",
        input="",
        output="",
        params=f" dist  {mash_sketch} {phrog_sketch} -d {mash_distance} -i ",
        logdir=logdir,
        outfile=mash_tsv,
    )

    # need to write to stdout

    try:
        ExternalTool.run_tool(mash_dist, to_stdout=True)

    except:
        logger.error("Error with mash dist\n")


def run_dnaapler(filepath_in, contig_count, out_dir, threads, logdir):
    """
    Runs dnaapler
    :param filepath_in: input fasta file
    :param contig_count: int numbher of contigs
    :param out_dir: output directory
    :param threads: number of threads
    :params prefix: prefix
    :return: dnaapler_success - whether dnaapler worked
    """

    logger.info(
        "Running Dnaapler to rerorient all contigs to begin with the terminase large subunit."
    )

    dnaapler_output_dir = os.path.join(out_dir, "dnaapler")

    if contig_count == 1:
        dnaapler = ExternalTool(
            tool="dnaapler phage",
            input=f"-i {filepath_in}",
            output=f"-o {dnaapler_output_dir} -t {threads}",
            params=f"",  # need strange order for minced params go first
            logdir=logdir,
            outfile="",
        )
    else:
        dnaapler = ExternalTool(
            tool="dnaapler all",
            input=f"-i {filepath_in}",
            output=f"-o {dnaapler_output_dir} -t {threads}",
            params=f"",  # need strange order for minced params go first
            logdir=logdir,
            outfile="",
        )

    dnaapler_success = True
    try:
        ExternalTool.run_tool(dnaapler)
    except:
        logger.warning("Dnaapler failed to find the large terminase subunit.")
        logger.warning(
            "For reorientation, please run pharokka again without --dnaapler and use --terminase instead."
        )
        logger.warning("Pharokka will now continue without reorientation.")
        dnaapler_success = False

    return dnaapler_success


# #### skani

# # no need to sketch
# # takes like 10 secs to run this

# # pharokka_v1.4.0_databases % skani dist -q ../pharokka/tests/test_data/overall/Meta_example/combined_meta.fasta -r 1Jan2024_genomes.fa  --qi --ri --slow -m 200 -n 1  > out


# def run_skani(input_fasta, out_dir, db_dir, logdir):
#     """
#     Runs mash
#     :param filepath_in: input filepath
#     :param out_dir: output directory
#     :param mash_distance: mash distance - float
#     :param logger logger
#     :return:
#     """

#     inphared_fasta = os.path.join(db_dir, "1Jan2024_genomes.fa.gz")
#     skani_tsv = os.path.join(out_dir, "skani_out.tsv")

#     mash_dist = ExternalTool(
#         tool="skani",
#         input="",
#         output="",
#         params=f" dist  -q {input_fasta} -r {inphared_fasta} --qi --ri --slow -m 200 -n 1 ",
#         logdir=logdir,
#         outfile=skani_tsv,
#     )

#     # need to write to stdout

#     try:
#         ExternalTool.run_tool(mash_dist, to_stdout=True)

#     except:
#         logger.error("Error with mash dist\n")



#!/usr/bin/env python3
import argparse
import os
import sys
from argparse import RawTextHelpFormatter
from pathlib import Path

from input_commands import validate_fasta
from loguru import logger
from plot import create_plot
from util import get_version


def get_input():
    """gets input for pharokka_plotter.py
    :return: args
    """
    parser = argparse.ArgumentParser(
        description="pharokka_plotter.py: pharokka plotting function",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--infile",
        action="store",
        required=True,
        help="Input genome file in FASTA format.",
    )
    parser.add_argument(
        "-n",
        "--plot_name",
        action="store",
        default="pharokka_plot",
        help='Output plot file name. ".png" suffix will be added to this automatically. \nWill be output in the Pharokka output directory if -o is specified, or in the working directory if --gff andf --genbank are specified.',
    )
    parser.add_argument(
        "-o", "--outdir", action="store", default="", help="Pharokka output directory."
    )
    parser.add_argument("--gff", action="store", default="", help="Pharokka gff.")
    parser.add_argument(
        "--genbank", action="store", default="", help="Pharokka genbank."
    )
    parser.add_argument(
        "-p",
        "--prefix",
        action="store",
        help="Prefix used to create pharokka output. Will default to pharokka.",
        default="pharokka",
    )
    parser.add_argument(
        "-t", "--plot_title", action="store", default="Phage", help="Plot name."
    )
    parser.add_argument(
        "-f", "--force", help="Overwrites the output plot file.", action="store_true"
    )
    parser.add_argument(
        "--label_hypotheticals",
        help="Flag to label  hypothetical or unknown proteins. By default these are not labelled.",
        action="store_true",
    )
    parser.add_argument(
        "--remove_other_features_labels",
        help="Flag to remove labels for tRNA/tmRNA/CRISPRs. By default these are labelled. \nThey will still be plotted in black.",
        action="store_true",
    )
    parser.add_argument(
        "--title_size",
        action="store",
        default="20",
        help="Controls title size. Must be an integer. Defaults to 20.",
    )
    parser.add_argument(
        "--label_size",
        action="store",
        default="8",
        help="Controls annotation label size. Must be an integer. Defaults to 8.",
    )
    parser.add_argument(
        "--interval",
        action="store",
        default="5000",
        help="Axis tick interval. Must be an integer. Must be an integer. Defaults to 5000.",
    )
    parser.add_argument(
        "--truncate",
        action="store",
        default="20",
        help="Number of characters to include in annoation labels before truncation with ellipsis. \nMust be an integer. Defaults to 20.",
    )
    parser.add_argument(
        "--dpi",
        action="store",
        default="600",
        help="Resultion (dots per inch). Must be an integer. Defaults to 600.",
    )
    parser.add_argument(
        "--annotations",
        action="store",
        default="1",
        help="Controls the proporition of annotations labelled. Must be a number between 0 and 1 inclusive. \n0 = no annotations, 0.5 = half of the annotations, 1 = all annotations. \nDefaults to 1. Chosen in order of CDS size.",
    )
    parser.add_argument(
        "--label_ids",
        action="store",
        default="",
        help="Text file with list of CDS IDs (from gff file) that are guaranteed to be labelled.",
    )
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = get_input()
    # preamble
    logger.info(f"Starting Pharokka v{get_version()}")
    logger.info("Running pharokka_plotter.py to plot your phage.")
    logger.info("Command executed: {}", args)
    logger.info("Repository homepage is https://github.com/gbouras13/pharokka")
    logger.info("Written by George Bouras: george.bouras@adelaide.edu.au")
    logger.info("Checking your inputs.")
    logger.add(lambda _: sys.exit(1), level="ERROR") # https://github.com/gbouras13/plassembler/pull/69
    try:
        int(args.interval)
    except:
        logger.error(
            f"--interval {args.interval} specified is not an integer. Please check your input and try again."
        )

    try:
        int(args.label_size)
    except:
        logger.error(
            f"--label_size {args.label_size} specified is not an integer. Please check your input and try again."
        )

    try:
        int(args.title_size)
    except:
        logger.error(
            f"--title_size {args.title_size} specified is not an integer. Please check your input and try again."
        )

    try:
        int(args.dpi)
    except:
        logger.error(
            f"--dpi {args.dpi} specified is not an integer. Please check your input and try again."
        )

    try:
        float(args.annotations)
    except:
        logger.error(
            f"--annotations {args.annotations} specified is not an float. Please check your input and try again."
        )

    # check if plot already exists
    if args.outdir == "":
        plot_file = str(args.plot_name) + ".png"
        svg_plot_file = str(args.plot_name) + ".svg"
    else:
        plot_file = os.path.join(args.outdir, f"{args.plot_name}.png")
        svg_plot_file = os.path.join(args.outdir, f"{args.plot_name}.svg")

    if args.force == True:
        if os.path.isfile(plot_file) == True:
            os.remove(plot_file)
        else:
            logger.warning(
                "--force was specified even though the output plot file does not already exist."
            )
            logger.warning("Continuing")
    else:
        if os.path.isfile(plot_file) == True:
            logger.error(
                f"Output plot file {plot_file} already exists and force was not specified. Please specify -f or --force to overwrite the output plot file."
            )

    if args.force == True:
        if os.path.isfile(svg_plot_file) == True:
            os.remove(svg_plot_file)
        else:
            logger.warning(
                "--force was specified even though the output plot file does not already exist."
            )
            logger.warning("Continuing")
    else:
        if os.path.isfile(svg_plot_file) == True:
            logger.error(
                f"Output plot file {svg_plot_file} already exists and force was not specified. Please specify -f or --force to overwrite the output plot file."
            )

    # flag to see if user provided gff and genbank or output directory
    gff_genbank_flag = True

    if args.gff == "" or args.genbank == "":
        if args.outdir == "":
            logger.error(
                "You have not specified both a Pharokka gff and a gbk file, or a Pharokka output directory. Please check your input and try again."
            )
        else:
            logger.info("You have specified a Pharokka output directory.")
            gff_genbank_flag = False

            # check the outdir exists
            if os.path.isdir(args.outdir) == False:
                logger.error(
                    f"Provided Pharokka output directory {args.outdir} does not exist. Please check -o/--outdir."
                )

    else:
        if args.outdir != "":
            logger.info(
                "You have specified Pharokka genbank and gff files and also an output directory."
            )
            logger.info(
                "The directory will be ignored. Continuing with the gff and genbank file."
            )
        else:
            logger.info("You have specified both Pharokka genbank and gff files.")

    # check the input fasta exists
    validate_fasta(args.infile)

    # get num input contigs
    num_contigs = len([1 for line in open(args.infile) if line.startswith(">")])
    if num_contigs > 1:
        logger.warning(
            "More than one contig detected in the input file. Only the first phage contig will be plotted."
        )

    # gff file

    if gff_genbank_flag == True:
        gff_file = args.gff
    else:
        gff_file = os.path.join(args.outdir, args.prefix + ".gff")

    if os.path.isfile(gff_file) == False:
        logger.error(
            f" {gff_file} was not found. Please check the prefix value `-p` matches the pharokka output directory, \n or use --gff and --genbank to specify the gff and genbank files and try again."
        )

    # Load Genbank file
    gbk_file = os.path.join(args.outdir, args.prefix + ".gbk")

    if gff_genbank_flag == True:
        gbk_file = args.genbank
    else:
        gbk_file = os.path.join(args.outdir, args.prefix + ".gbk")

    # validate gbk exists

    if os.path.isfile(gbk_file) == False:
        logger.error(
            f" {gbk_file} was not found. Please check the prefix value `-p` matches the pharokka output directory, \n or use --gff and --genbank to specify the gff and genbank files and try again."
        )

    # check label_ids

    # list of all IDs that need to be labelled from file
    label_force_list = []

    if args.label_ids != "":
        logger.info(
            f"You have specified a file {args.label_ids} containing a list of CDS IDs to force label."
        )
        # check if it is a file
        if os.path.isfile(args.label_ids) == False:
            logger.error(f"{args.label_ids} was not found.")
        # check if it contains text
        try:
            # Open the file in read mode
            with open(Path(args.label_ids), "r") as file:
                # Read the first character
                first_char = file.read(1)

                # read all the labels
                with open(Path(args.label_ids)) as f:
                    ignore_dict = {x.rstrip().split()[0] for x in f}
                # label force list
                label_force_list = list(ignore_dict)

        except FileNotFoundError:
            logger.warning(
                f"{args.label_id} contains no text. No contigs will be ignored"
            )

    logger.info("All files checked.")
    logger.info("Plotting the phage.")

    create_plot(
        gff_file,
        gbk_file,
        args.interval,
        args.annotations,
        args.title_size,
        args.plot_title,
        args.truncate,
        plot_file,
        svg_plot_file,
        args.dpi,
        args.label_size,
        args.label_hypotheticals,
        args.remove_other_features_labels,
        label_force_list,
    )



import os

import pyrodigal_gv
from Bio import SeqIO

# from Bio.Seq import Seq
# from Bio.SeqRecord import SeqRecord
# from external_tools import ExternalTool
# from loguru import logger
# from util import remove_directory


def run_pyrodiga_gv(filepath_in, out_dir, coding_table):
    """
    Gets CDS using pyrodigal_gv
    :param filepath_in: input filepath
    :param out_dir: output directory
    :param logger logger
    :param meta Boolean - metagenomic mode flag
    :param coding_table coding table for prodigal (default 11)
    :return:
    """

    # true
    orf_finder = pyrodigal_gv.ViralGeneFinder(meta=True)

    with open(os.path.join(out_dir, "prodigal_out.gff"), "w") as dst:
        for i, record in enumerate(SeqIO.parse(filepath_in, "fasta")):
            genes = orf_finder.find_genes(str(record.seq))
            genes.write_gff(dst, sequence_id=record.id)

    with open(os.path.join(out_dir, "prodigal_out_tmp.fasta"), "w") as dst:
        for i, record in enumerate(SeqIO.parse(filepath_in, "fasta")):
            genes = orf_finder.find_genes(str(record.seq))
            genes.write_genes(dst, sequence_id=record.id)



#!/usr/bin/env python3

import os
import shutil
import sys
import time
import warnings
from pathlib import Path

from Bio import BiopythonDeprecationWarning
from custom_db import run_custom_pyhmmer
from databases import check_db_installation
from hmm import run_pyhmmer
from input_commands import (check_dependencies, get_input, instantiate_dirs,
                            instantiate_split_output,
                            validate_and_extract_genbank, validate_custom_hmm,
                            validate_fasta, validate_gene_predictor,
                            validate_meta, validate_terminase,
                            validate_threads)
from loguru import logger
from post_processing import Pharok, remove_post_processing_files
from processes import (concat_phanotate_meta, concat_trnascan_meta,
                       convert_gff_to_gbk, reorient_terminase, run_aragorn,
                       run_dnaapler, run_mash_dist, run_mash_sketch,
                       run_minced, run_mmseqs, run_phanotate,
                       run_phanotate_fasta_meta, run_phanotate_txt_meta,
                       run_pyrodigal, run_pyrodigal_gv, run_pyrodigal_rv, run_trna_scan,
                       run_trnascan_meta, split_input_fasta, translate_fastas)
from util import count_contigs, get_version

# add this to make sure of deprecation warning with biopython
warnings.simplefilter("default", BiopythonDeprecationWarning)


def main():
    # get the args
    args = get_input()

    logger.add(lambda _: sys.exit(1), level="ERROR") # https://github.com/gbouras13/plassembler/pull/69


    if args.citation == True:
        logger.info("If you use Pharokka in your research, please cite:")
        logger.info(
            "George Bouras, Roshan Nepal, Ghais Houtak, Alkis James Psaltis, Peter-John Wormald, Sarah Vreugde."
        )
        logger.info("Pharokka: a fast scalable bacteriophage annotation tool.")
        logger.info("Bioinformatics, Volume 39, Issue 1, January 2023, btac776.")
        logger.info("https://doi.org/10.1093/bioinformatics/btac776.")
        logger.info(
            "You should also cite the full list of tools Pharokka uses, which can be found at https://github.com/gbouras13/pharokka#citation."
        )
        sys.exit()

    # set the prefix
    if args.prefix == "Default":
        prefix = "pharokka"
    else:
        prefix = args.prefix

    # set the locustag flag for locustag generation
    if args.locustag == "Default":
        locustag = "Random"
    else:
        locustag = args.locustag

    # set the gene_predictor
    gene_predictor = args.gene_predictor
    # set to phanotate by default and prodigal-gv in meta mode
    if gene_predictor == "default":
        if args.meta is True:
            gene_predictor = "prodigal-gv"
        else:
            gene_predictor = "phanotate"

    # instantiate outdir
    out_dir = instantiate_dirs(args.outdir, args.meta, args.force)

    # set log dir
    logdir = Path(f"{out_dir}/logs")

    # set the db dir
    if args.database == "Default":
        db_dir = os.path.join(os.path.realpath(__file__), "../", "databases/")
    else:
        db_dir = args.database

    # get start time
    start_time = time.time()
    # initial logging stuff
    log_file = os.path.join(args.outdir, f"pharokka_{start_time}.log")
    # adds log file
    logger.add(log_file)
    # preamble
    logger.info(f"Starting Pharokka v{get_version()}")
    logger.info("Command executed: {}", args)
    logger.info("Repository homepage is https://github.com/gbouras13/pharokka")
    logger.info("Written by George Bouras: george.bouras@adelaide.edu.au")

    logger.info(f"Checking database installation in {db_dir}.")
    database_installed = check_db_installation(db_dir)
    if database_installed == True:
        logger.info("All databases have been successfully checked.")
    else:
        logger.error(
            "The database directory was unsuccessfully checked. Please run install_databases.py."
        )

    ### custom hmm

    custom_hmm_flag = False
    if args.custom_hmm != "":
        custom_hmm_flag = True

    if custom_hmm_flag is True:
        validate_custom_hmm(args.custom_hmm)

    # dependencies
    logger.info("Checking dependencies.")
    # output versions
    (
        phanotate_version,
        pyrodigal_version,
        pyrodigal_gv_version,
        pyrodigal_rv_version,
        trna_version,
        aragorn_version,
        minced_version,
    ) = check_dependencies(args.skip_mash)

    # instantiation/checking fasta and gene_predictor
    if args.genbank is True:
        logger.info("You have specified --genbank.")
        logger.info(
            f"Therefore, {args.infile} is a genbank file instead of a FASTA file."
        )
        logger.info("Your custom CDS calls in this genbank file will be preserved.")
        validate_and_extract_genbank(args.infile, out_dir)
        gene_predictor = "genbank"
        input_fasta = f"{out_dir}/genbank.fasta"
    else:
        validate_fasta(args.infile)
        input_fasta = args.infile

    # other validations
    validate_gene_predictor(gene_predictor, args.genbank)
    validate_threads(args.threads)

    ###################
    # define input
    ###################

    # reorient with dnaapler if chosen
    if args.dnaapler is True:
        logger.info(
            "You have chosen to reorient your contigs by specifying --dnaapler. Checking the input."
        )

        # in case both --dnaapler and --terminase is selected
        if args.terminase == True:
            logger.info("Ignoring --terminase. Dnaapler will be run instead.")
            args.terminase = False
            args.terminase_strand = "nothing"
            args.terminase_start = "nothing"

        # count contigs
        contig_count = count_contigs(input_fasta)

        # runs dnaapler
        dnaapler_success = run_dnaapler(
            input_fasta, contig_count, out_dir, args.threads, logdir
        )

        if dnaapler_success == True:
            input_fasta = os.path.join(out_dir, "dnaapler/dnaapler_reoriented.fasta")
            destination_file = os.path.join(
                out_dir, f"{prefix}_dnaapler_reoriented.fasta"
            )
            # copy FASTA to output
            shutil.copy(input_fasta, destination_file)

    # terminase reorienting if chosen and dnaapler is False
    if args.dnaapler is False:
        if args.terminase is True:
            logger.info(
                "You have chosen to reorient your genome by specifying --terminase. Checking the input."
            )
            validate_terminase(input_fasta, args.terminase_strand, args.terminase_start)
            reorient_terminase(
                input_fasta,
                out_dir,
                prefix,
                args.terminase_strand,
                args.terminase_start,
            )
            # overwrite input_fasta if terminase reorienter is true
            input_fasta = os.path.join(
                out_dir, prefix + "_genome_terminase_reoriented.fasta"
            )

        if args.terminase_strand != "nothing" and args.terminase == False:
            logger.info(
                "You have specified a terminase strand using --strand to reorient your genome, but you have not specified --terminase to activate terminase mode. \nContinuing without reorientation."
            )

        if args.terminase_strand != "nothing" and args.terminase_start == False:
            logger.info(
                "You have specified a terminase start coordinate using --terminase_start to reorient your genome, but you have not specified --terminase to activate terminase mode. \nContinuing without reorientation."
            )

    ########
    # mmseqs2 and hmm decisions
    ########

    # can't have fast and mmseqs2 only
    if args.fast is True and args.mmseqs2_only is True:
        logger.error(
            "You have specified --fast or --hmm_only and --mmseqs2_only. This is impossible. Please choose one or the other."
        )

    # can't have fast and meta_hmm
    if args.fast is True and args.meta_hmm is True:
        logger.error(
            "You have specified --fast or --hmm_only and --meta_hmm. This is impossible. Please choose one or the other."
        )

    # can't have mmseqs2 only and meta_hmm
    if args.mmseqs2_only is True and args.meta_hmm is True:
        logger.error(
            "You have specified --mmseqs2_only and --meta_hmm. This is impossible. Please choose one or the other."
        )

    # by default run both not in meta mode
    mmseqs_flag = True
    hmm_flag = True

    if args.meta is True:  # meta mode default only mmseqs
        if args.meta_hmm is True:  # with --meta_hmm
            logger.info(
                "You have specified --meta_hmm and -m/--meta to run PyHMMER as well as MMseqs2 in meta mode. This may be take a while, please be patient."
            )
            mmseqs_flag = True
            hmm_flag = True
        else:  # just mmseqs2 by default
            mmseqs_flag = True
            hmm_flag = False
    else:  # not in meta mode
        if args.meta_hmm is True:
            logger.warning(
                "You have specified --meta_hmm to run PyHMMER in meta mode, but you have not specified -m to activate meta mode."
            )
            logger.warning("Ignoring --meta_hmm.")

    # overrides if fast/hmm_only is chosen
    if args.fast == True:
        logger.info("You have specified --fast or --hmm_only. MMseqs2 will not be run.")
        if args.meta is True:
            logger.warning(
                "You have specified --fast with -m/--meta. This may be take a while, please be patient."
            )
        mmseqs_flag = False
        hmm_flag = True

    if args.mmseqs2_only == True:
        logger.info("You have specified --mmseqs2_only. PyHMMER will not be run.")
        mmseqs_flag = True
        hmm_flag = False

    # validates meta mode
    validate_meta(input_fasta, args.meta, args.split, args.genbank)

    # meta mode split input for trnascan and maybe phanotate
    if args.meta == True:
        num_fastas = split_input_fasta(input_fasta, out_dir)
        # will generate split output gffs if split flag is true
        instantiate_split_output(out_dir, args.split)

    # CDS predicton
    # phanotate
    if gene_predictor == "phanotate":
        logger.info("Running Phanotate.")
        if args.meta == True:
            logger.info("Applying meta mode.")
            run_phanotate_fasta_meta(input_fasta, out_dir, args.threads, num_fastas)
            run_phanotate_txt_meta(input_fasta, out_dir, args.threads, num_fastas)
            concat_phanotate_meta(out_dir, num_fastas)
        else:
            run_phanotate(input_fasta, out_dir, logdir)
    elif gene_predictor == "prodigal":
        logger.info("Implementing Prodigal using Pyrodigal.")
        transl_table = run_pyrodigal(
            input_fasta, out_dir, args.meta, args.coding_table, int(args.threads)
        )
        args.coding_table = transl_table
    elif gene_predictor == "genbank":
        logger.info("Extracting CDS information from your genbank file.")
    elif gene_predictor == "prodigal-gv":
        logger.info("Implementing Prodigal-gv using Pyrodigal-gv.")
        run_pyrodigal_gv(input_fasta, out_dir, int(args.threads))
    elif gene_predictor == "pyrodigal-rv":
        logger.info("Implementing Pyrodigal-rv.")
        run_pyrodigal_rv(input_fasta, out_dir, int(args.threads))

    # translate fastas (parse genbank) - gets the required CDS if --genbank is used too
    translate_fastas(out_dir, gene_predictor, args.coding_table, args.infile)

    # run trna-scan meta mode if required
    if args.skip_extra_annotations is False:
        if args.meta == True:
            logger.info("Starting tRNA-scanSE. Applying meta mode.")
            run_trnascan_meta(
                input_fasta, out_dir, args.threads, num_fastas, args.trna_scan_model
            )
            concat_trnascan_meta(out_dir, num_fastas)
        else:
            logger.info("Starting tRNA-scanSE.")
            run_trna_scan(
                input_fasta, args.threads, out_dir, logdir, args.trna_scan_model
            )
        # run minced and aragorn
        run_minced(input_fasta, out_dir, prefix, args.minced_args, logdir)
        run_aragorn(input_fasta, out_dir, prefix, logdir)

    # running mmseqs2 on the 3 databases
    if mmseqs_flag is True:
        logger.info("Starting MMseqs2.")
        run_mmseqs(
            db_dir,
            out_dir,
            args.threads,
            logdir,
            gene_predictor,
            args.evalue,
            args.reverse_mmseqs2,
            args.sensitivity,
            db_name="PHROG"
        )
        run_mmseqs(
            db_dir,
            out_dir,
            args.threads,
            logdir,
            gene_predictor,
            args.evalue,
            args.reverse_mmseqs2,
            args.sensitivity,
            db_name="CARD"
        )
        run_mmseqs(
            db_dir,
            out_dir,
            args.threads,
            logdir,
            gene_predictor,
            args.evalue,
            args.reverse_mmseqs2,
            args.sensitivity,
            db_name="VFDB"
        )

    if hmm_flag is True:
        # runs pyhmmer on PHROGs
        logger.info("Running PyHMMER on PHROGs.")
        best_results_pyhmmer = run_pyhmmer(
            db_dir, out_dir, args.threads, gene_predictor, args.evalue
        )

    if custom_hmm_flag is True:
        # runs pyhmmer on customer
        logger.info(f"Running PyHMMER on custom HMM database {args.custom_hmm}.")
        best_results_custom_pyhmmer = run_custom_pyhmmer(
            args.custom_hmm, out_dir, args.threads, gene_predictor, args.evalue
        )

    #################################################
    # post processing
    #################################################

    logger.info("Post Processing Output.")

    # instanatiate the class with some of the params
    pharok = Pharok()
    pharok.out_dir = out_dir
    pharok.db_dir = db_dir
    pharok.gene_predictor = gene_predictor
    pharok.prefix = prefix
    pharok.locustag = locustag
    pharok.input_fasta = input_fasta
    pharok.meta_mode = args.meta
    pharok.coding_table = args.coding_table
    pharok.mmseqs_flag = mmseqs_flag
    pharok.hmm_flag = hmm_flag
    pharok.custom_hmm_flag = custom_hmm_flag
    pharok.phanotate_version = phanotate_version
    pharok.pyrodigal_version = pyrodigal_version
    pharok.pyrodigal_gv_version = pyrodigal_gv_version
    pharok.pyrodigal_rv_version = pyrodigal_rv_version
    pharok.trna_version = trna_version
    pharok.aragorn_version = aragorn_version
    pharok.minced_version = minced_version
    pharok.skip_extra_annotations = args.skip_extra_annotations
    pharok.reverse_mmseqs2 = args.reverse_mmseqs2

    if pharok.hmm_flag is True:
        pharok.pyhmmer_results_dict = best_results_pyhmmer
    if pharok.custom_hmm_flag is True:
        pharok.custom_pyhmmer_results_dict = best_results_custom_pyhmmer

    #####################################
    # post processing
    #####################################

    # gets df of length and gc for each contig
    pharok.get_contig_name_lengths()

    # post process results
    # includes vfdb and card
    # adds the merged df, vfdb and card top hits dfs to the class objec
    # no need to specify params as they are in the class :)
    pharok.process_results()

    # parse the aragorn output
    # only if not skipping annots
    if args.skip_extra_annotations is False:
        pharok.parse_aragorn()

    # create gff and save locustag to class for table
    pharok.create_gff()

    # create table
    pharok.create_tbl()

    # output single gffs in meta mode
    if args.split == True and args.meta == True:
        # create gffs for each contig
        pharok.create_gff_singles()
        # converts each gff to gbk
        pharok.convert_singles_gff_to_gbk()
        # splits the input fasta into single fastas
        pharok.split_fasta_singles()

    # create and write vfdb and card tophits
    # needs to be before .create_txt or else won't count properly
    # will only write to file if mmseqs2 has been run (in function)

    pharok.write_tophits_vfdb_card()

    # write the summary tsv outputs and functions
    pharok.create_txt()

    # convert to genbank
    logger.info("Converting gff to genbank.")
    # not part of the class so from processes.py
    convert_gff_to_gbk(input_fasta, out_dir, out_dir, prefix, pharok.prot_seq_df)

    # update fasta headers and final output tsv
    pharok.update_fasta_headers()
    pharok.update_final_output()

    # output single gffs in meta mode
    if args.split == True and args.meta == True:
        # splits the faa into single .faa
        pharok.split_faas_singles()

    # extract terL
    pharok.extract_terl()

    # run mash
    if args.skip_mash is False:  # skips mash
        logger.info("Finding the closest match for each contig in INPHARED using mash.")
        # in process.py
        run_mash_sketch(input_fasta, out_dir, logdir)
        run_mash_dist(out_dir, db_dir, args.mash_distance, logdir)
        # part of the class
        pharok.inphared_top_hits()
    else:
        logger.info("You have chosen --skip_mash.")
        logger.info(
            "Skipping finding the closest match for each contig in INPHARED using mash."
        )

    # delete tmp files
    remove_post_processing_files(
        out_dir, gene_predictor, args.meta, args.keep_raw_prodigal
    )

    # Determine elapsed time
    elapsed_time = time.time() - start_time
    elapsed_time = round(elapsed_time, 2)

    # Show elapsed time for the process
    logger.info("Pharokka has finished.")
    logger.info(f"Elapsed time: {elapsed_time} seconds")

    logger.info("If you use Pharokka in your research, please cite:")
    logger.info(
        "George Bouras, Roshan Nepal, Ghais Houtak, Alkis James Psaltis, Peter-John Wormald, Sarah Vreugde."
    )
    logger.info("Pharokka: a fast scalable bacteriophage annotation tool.")
    logger.info("Bioinformatics, Volume 39, Issue 1, January 2023, btac776.")
    logger.info("https://doi.org/10.1093/bioinformatics/btac776.")
    logger.info(
        "You should also cite the full list of tools Pharokka uses, which can be found at https://github.com/gbouras13/pharokka#citation."
    )


if __name__ == "__main__":
    main()



import hashlib
import shlex
import subprocess
import sys
from pathlib import Path
from typing import List, Optional, Tuple

from loguru import logger

"""
taken from tbpore https://github.com/mbhall88/tbpore/blob/main/tbpore/external_tools.py

"""


class ExternalTool:
    def __init__(
        self,
        tool: str,
        input: str,
        output: str,
        params: str,
        logdir: Path,
        outfile: Path,
    ):
        self.command: List[str] = self._build_command(tool, input, output, params)
        logdir.mkdir(parents=True, exist_ok=True)
        command_hash = hashlib.sha256(self.command_as_str.encode("utf-8")).hexdigest()
        tool_name = Path(tool).name
        logfile_prefix: Path = logdir / f"{tool_name}_{command_hash}"
        self.out_log = f"{logfile_prefix}.out"
        self.err_log = f"{logfile_prefix}.err"
        self.outfile = outfile

    @property
    def command_as_str(self) -> str:
        return shlex.join(self.command)

    @staticmethod
    def _build_command(tool: str, input: str, output: str, params: str) -> List[str]:
        # note: shlex.join does not allow us to shlex.split() later
        # this is explicitly a " ".join()
        command = " ".join([tool, params, output, input])
        escaped_command = shlex.split(command)
        return escaped_command

    def run(self) -> None:
        with open(self.out_log, "w") as stdout_fh, open(self.err_log, "w") as stderr_fh:
            print(f"Command line: {self.command_as_str}", file=stderr_fh)
            logger.info(f"Started running {self.command_as_str} ...")
            self._run_core(self.command, stdout_fh=stdout_fh, stderr_fh=stderr_fh)
            logger.info(f"Done running {self.command_as_str}")

    def run_to_stdout(
        self,
    ) -> None:
        with open(self.outfile, "w") as outfile, open(self.err_log, "w") as stderr_fh:
            print(f"Command line: {self.command_as_str}", file=stderr_fh)
            logger.info(f"Started running {self.command_as_str} ...")
            self._run_core(self.command, stdout_fh=outfile, stderr_fh=stderr_fh)
            logger.info(f"Done running {self.command_as_str}")

    @staticmethod
    def _run_core(command: List[str], stdout_fh, stderr_fh) -> None:
        subprocess.check_call(command, stdout=stdout_fh, stderr=stderr_fh)

    @staticmethod
    def run_tools(tools_to_run: Tuple["ExternalTool", ...]) -> None:
        for tool in tools_to_run:
            try:
                tool.run()
            except subprocess.CalledProcessError as error:
                logger.error(
                    f"Error calling {tool.command_as_str} (return code {error.returncode})"
                )
                logger.error(f"Please check stdout log file: {tool.out_log}")
                logger.error(f"Please check stderr log file: {tool.err_log}")
                logger.error("Temporary files are preserved for debugging")
                logger.error("Exiting...")
                sys.exit(1)

    """
    Only one toolf
    """

    @staticmethod
    def run_tool(tool: "ExternalTool", to_stdout: Optional[bool] = False) -> None:
        try:
            if to_stdout is False:
                tool.run()
            else:  # if the tool needs to run to stdout
                tool.run_to_stdout()
        except subprocess.CalledProcessError as error:
            logger.error(
                f"Error calling {tool.command_as_str} (return code {error.returncode})"
            )
            logger.error(f"Please check stdout log file: {tool.out_log}")
            logger.error(f"Please check stderr log file: {tool.err_log}")
            logger.error("Temporary files are preserved for debugging")
            logger.error("Exiting...")

            sys.exit(1)



#!/usr/bin/env python3

import os
import sys
import time
from pathlib import Path

from databases import check_db_installation
from input_commands import (check_dependencies, instantiate_dirs,
                            validate_fasta, validate_threads)
from loguru import logger
from post_processing import remove_directory, remove_file
from proteins import (Pharok_Prot, get_input_proteins, run_mmseqs_proteins,
                      run_pyhmmer_proteins)
from util import get_version


def main():
    # get the args
    args = get_input_proteins()

    logger.add(lambda _: sys.exit(1), level="ERROR") # https://github.com/gbouras13/plassembler/pull/69


    if args.citation == True:
        logger.info("If you use Pharokka in your research, please cite:")
        logger.info(
            "George Bouras, Roshan Nepal, Ghais Houtak, Alkis James Psaltis, Peter-John Wormald, Sarah Vreugde."
        )
        logger.info("Pharokka: a fast scalable bacteriophage annotation tool.")
        logger.info("Bioinformatics, Volume 39, Issue 1, January 2023, btac776.")
        logger.info("https://doi.org/10.1093/bioinformatics/btac776.")
        logger.info(
            "You should also cite the full list of tools Pharokka uses, which can be found at https://github.com/gbouras13/pharokka#citation."
        )
        sys.exit()

    # set the prefix
    if args.prefix == "Default":
        prefix = "pharokka_proteins"
    else:
        prefix = args.prefix

    # instantiate outdir
    out_dir = instantiate_dirs(args.outdir, False, args.force)  # args.meta always false

    # set log dir
    logdir = Path(f"{out_dir}/logs")

    # set the db dir
    if args.database == "Default":
        db_dir = os.path.join(os.path.realpath(__file__), "../", "databases/")
    else:
        db_dir = args.database

    # get start time
    start_time = time.time()
    # initial logging stuff
    log_file = os.path.join(args.outdir, f"pharokka_proteins_{start_time}.log")
    # adds log file
    logger.add(log_file)
    
    # preamble
    logger.info(f"Starting Pharokka v{get_version()}")
    logger.info("Running pharokka_proteins.py to annotate proteins.")
    logger.info("Command executed: {}", args)
    logger.info("Repository homepage is https://github.com/gbouras13/pharokka")
    logger.info("Written by George Bouras: george.bouras@adelaide.edu.au")

    # check the database is installed
    logger.info("Checking database installation.")

    database_installed = check_db_installation(db_dir)
    if database_installed == True:
        logger.info("All databases have been successfully checked.")
    else:
        logger.error(
            "The database directory was unsuccessfully checked. Please run install_databases.py"
        )

    # dependencies
    logger.info("Checking dependencies.")
    check_dependencies(False)  # to check pharokka_proteins.py, don't need mash

    # instantiation/checking fasta
    validate_fasta(args.infile)
    validate_threads(args.threads)

    # define input - overwrite if terminase reorienter is true
    input_fasta = args.infile

    ########
    # mmseqs2 and hmm decisions
    ########

    # can't have fast and mmseqs2 only
    if args.hmm_only is True and args.mmseqs2_only is True:
        logger.error(
            "You have specified --fast or --hmm_only and --mmseqs2_only. This is impossible. Please choose one or the other."
        )

    # by default
    mmseqs_flag = True
    hmm_flag = True

    if args.hmm_only is True:
        logger.info("You have specified --hmm_only. MMseqs2 will not be run.")
        mmseqs_flag = False
    if args.mmseqs2_only is True:
        logger.info("You have specified --mmseqs2_only. Pyhmmer will not be run.")
        hmm_flag = False

    # running mmseqs2 on the 3 databases
    if mmseqs_flag is True:
        logger.info("Starting MMseqs2.")
        run_mmseqs_proteins(
            input_fasta,
            db_dir,
            out_dir,
            args.threads,
            logdir,
            args.evalue,
            args.reverse_mmseqs2,
            args.sensitivity,
            db_name="PHROG",
        )
        run_mmseqs_proteins(
            input_fasta,
            db_dir,
            out_dir,
            args.threads,
            logdir,
            args.evalue,
            args.reverse_mmseqs2,
            args.sensitivity,
            db_name="CARD",
        )
        run_mmseqs_proteins(
            input_fasta,
            db_dir,
            out_dir,
            args.threads,
            logdir,
            args.evalue,
            args.reverse_mmseqs2,
            args.sensitivity,
            db_name="VFDB",
        )

    if hmm_flag is True:
        # runs pyhmmer on PHROGs
        logger.info("Running PyHMMER.")
        best_results_pyhmmer = run_pyhmmer_proteins(
            input_fasta, db_dir, args.threads, args.evalue
        )

    #################################################
    # post processing
    #################################################

    logger.info("Post Processing Output.")

    # instanatiate the class with some of the params
    pharok = Pharok_Prot()
    pharok.out_dir = out_dir
    pharok.db_dir = db_dir
    pharok.prefix = prefix
    pharok.input_fasta = input_fasta
    pharok.mmseqs_flag = mmseqs_flag
    pharok.reverse_mmseqs2 = args.reverse_mmseqs2
    pharok.hmm_flag = hmm_flag
    if pharok.hmm_flag is True:
        pharok.pyhmmer_results_dict = best_results_pyhmmer

    # post process results
    # includes vfdb and card
    # adds the merged df, vfdb and card top hits dfs to the class objec
    # no need to specify params as they are in the class :)
    pharok.process_dataframes()

    # updates fasta headers
    pharok.update_fasta_headers()

    # cleanup

    remove_directory(os.path.join(out_dir, "mmseqs"))
    remove_file(os.path.join(out_dir, "CARD_results.tsv"))
    remove_file(os.path.join(out_dir, "vfdb_results.tsv"))
    remove_directory(os.path.join(out_dir, "CARD"))
    remove_directory(os.path.join(out_dir, "vfdb"))
    remove_directory(os.path.join(out_dir, "VFDB_dir"))
    remove_directory(os.path.join(out_dir, "tmp_dir"))
    remove_file(os.path.join(out_dir, "mmseqs_results.tsv"))
    remove_file(os.path.join(out_dir, "CARD_results.tsv"))
    remove_file(os.path.join(out_dir, "vfdb_results.tsv"))
    remove_directory(os.path.join(out_dir, "CARD"))
    remove_directory(os.path.join(out_dir, "vfdb"))

    # Determine elapsed time
    elapsed_time = time.time() - start_time
    elapsed_time = round(elapsed_time, 2)

    # Show elapsed time for the process
    logger.info("Pharokka has finished")
    logger.info("Elapsed time: " + str(elapsed_time) + " seconds")

    logger.info("If you use Pharokka in your research, please cite:")
    logger.info(
        "George Bouras, Roshan Nepal, Ghais Houtak, Alkis James Psaltis, Peter-John Wormald, Sarah Vreugde."
    )
    logger.info("Pharokka: a fast scalable bacteriophage annotation tool.")
    logger.info("Bioinformatics, Volume 39, Issue 1, January 2023, btac776.")
    logger.info("https://doi.org/10.1093/bioinformatics/btac776.")
    logger.info(
        "You should also cite the full list of tools Pharokka uses, which can be found at https://github.com/gbouras13/pharokka#citation."
    )


if __name__ == "__main__":
    main()



"""
to tar DBs
export GZIP=-9

To remake the MMseqs2 PHROG profile DB ith MMseqs2 v18
wget https://phrogs.lmge.uca.fr/downloads_from_website/MSA_phrogs.tar.gz
tar -xzf MSA_phrogs.tar.gz
# to rename for the headers
for f in MSA_Phrogs_M50_FASTA/*.fma; do     mv "$f" "${f%.fma}"; done
tar -czf MSA_phrogs_renamed.tar.gz MSA_Phrogs_M50_FASTA/
mmseqs tar2db  MSA_phrogs_renamed.tar.gz  msa --output-dbtype 11
mmseqs msa2profile msa phrogs_profile_db


tar cvzf pharokka_v1.8.0_databases.tar.gz pharokka_v1.8.0_databases

to create the mmseqs2 databases (CARD)
with the latest download v 3.2.7 on August 21 2023
mmseqs createdb protein_fasta_protein_homolog_model.fasta CARD

# also need aro.tsv for this one

# for VFDB, only need the FASTA

# VFDB update as of August 18 2023 (not versioned)
# clustered

mmseqs easy-cluster VFDB_setB_pro_form.fas VFDBclusterRes tmp --min-seq-id 0.5 -c 0.8 --cov-mode 1
mmseqs createdb VFDBclusterRes_rep_seq.fasta vfdb

"""

#!/usr/bin/env python3
import hashlib
import os
import shutil
import tarfile
from pathlib import Path

import requests
from alive_progress import alive_bar
from loguru import logger
from post_processing import remove_directory

VERSION = "1.8.0"

# to hold information about the different DBs
VERSION_DICTIONARY = {
    "1.2.0": {
        "md5": "0014b7a982dbf071f8856a5a29a95e30",
        "major": 1,
        "minor": 2,
        "minorest": 0,
        "db_url": "https://zenodo.org/record/7563578/files/pharokka_v1.2.0_database.tar.gz",
        "dir_name": "pharokka_v1.2.0_databases",
        "inphared_mash": "5Jan2023_genomes.fa.msh",
        "inphared_annot": "5Jan2023_data.tsv",
    },
    "1.4.0": {
        "md5": "c21144209b993c06fae2dac906d73b96",
        "major": 1,
        "minor": 4,
        "minorest": 0,
        "db_url": "https://zenodo.org/record/8276347/files/pharokka_v1.4.0_databases.tar.gz",
        "dir_name": "pharokka_v1.4.0_databases",
        "inphared_mash": "1Aug2023_genomes.fa.msh",
        "inphared_annot": "1Aug2023_data.tsv",
    },
    "1.8.0": {
        "md5": "a63c485241b900a11989bd1821bfbb09",
        "major": 1,
        "minor": 8,
        "minorest": 0,
        "db_url": "https://zenodo.org/record/17110353/files/pharokka_v1.8.0_databases.tar.gz",
        "dir_name": "pharokka_v1.8.0_databases",
        "inphared_mash": "9Aug2025_genomes.fa.msh",
        "inphared_annot": "9Aug2025_data.tsv",
    }
}


PHROG_DB_NAMES = [
    "VERSION_1_8_0",
    "phrogs_profile_db",
    "phrogs_profile_db.dbtype",
    "phrogs_profile_db.index",
    "phrogs_profile_db.lookup",
    "phrogs_profile_db.source",
    "phrogs_profile_db_h",
    "phrogs_profile_db_h.dbtype",
    "phrogs_profile_db_h.index",
]

PHROG_HMM_NAMES = ["all_phrogs.h3m"]

VFDB_DB_NAMES = [
    "vfdb",
    "vfdb.dbtype",
    "vfdb.index",
    "vfdb.lookup",
    "vfdb.source",
    "vfdb_h",
    "vfdb_h.dbtype",
    "vfdb_h.index",
    "VFDBclusterRes_cluster.tsv",
    "VFDBclusterRes_rep_seq.fasta",
]

CARD_DB_NAMES = [
    "CARD",
    "CARD.dbtype",
    "CARD.index",
    "CARD.lookup",
    "CARD.source",
    "CARD_h",
    "CARD_h.dbtype",
    "CARD_h.index",
]


def instantiate_install(db_dir):
    instantiate_dir(db_dir)
    # check the database is installed
    logger.info(f"Checking Pharokka database installation in {db_dir}.")
    downloaded_flag = check_db_installation(db_dir)
    if downloaded_flag == True:
        logger.info("All Pharokka Databases have already been Downloaded and Checked.")
    else:
        logger.info("Some Databases are missing.")

        db_url = VERSION_DICTIONARY[VERSION]["db_url"]
        requiredmd5 = VERSION_DICTIONARY[VERSION]["md5"]

        logger.info(f"Downloading Pharokka Databases from {db_url}.")

        tarball_path = Path(f"{db_dir}/pharokka_v1.8.0_databases.tar.gz")

        download(db_url, tarball_path)

        # get_database_zenodo(db_dir)

        md5_sum = calc_md5_sum(tarball_path)

        if md5_sum == requiredmd5:
            logger.info(f"Database file download OK: {md5_sum}")
        else:
            logger.error(
                f"Error: corrupt database file! MD5 should be '{requiredmd5}' but is '{md5_sum}'"
            )

        logger.info(f"Extracting DB tarball: file={tarball_path}, output={db_dir}")
        untar(tarball_path, db_dir)
        tarball_path.unlink()

        # download the aro_index.tsv from pharokka github with fixed metadata
        # fixing issue #339
        # https://raw.github.com/gbouras13/pharokka/master/aro_index.tsv


        # logger.info(f"Downloading Updated CARD metadata")
        # requiredmd5_card = "d999d71b21bf13c21e57ec7591bd8a47"
        # aro_url = "https://raw.github.com/gbouras13/pharokka/master/aro_index.tsv"
        # aro_path =  Path(f"{db_dir}/aro_index.tsv")


        # download(aro_url, aro_path)

        # md5_sum_card = calc_md5_sum(aro_path)
        # if md5_sum_card == requiredmd5_card:
        #     logger.info(f"CARD metadata file download OK: {md5_sum_card}")
        # else:
        #     logger.error(
        #         f"Error: corrupt database file! MD5 should be '{requiredmd5_card}' but is '{md5_sum_card}'"
        #     )


"""
lots of this code from the marvellous bakta https://github.com/oschwengers/bakta, db.py specifically
"""


def download(db_url: str, tarball_path: Path):

    headers = {
        "User-Agent": f"pharokka/{VERSION} (contact: george.bouras@adelaide.edu.au)"
    }

    try:
        with tarball_path.open("wb") as fh_out, requests.get(
            db_url, stream=True, headers=headers
        ) as resp:
            total_length = resp.headers.get("content-length")
            if total_length is not None:  # content length header is set
                total_length = int(total_length)
            with alive_bar(total=total_length, scale="SI") as bar:
                for data in resp.iter_content(chunk_size=1024 * 1024):
                    fh_out.write(data)
                    bar(count=len(data))
    except IOError:
        logger.error(
            f"ERROR: Could not download file from Zenodo! url={db_url}, path={tarball_path}"
        )


def calc_md5_sum(tarball_path: Path, buffer_size: int = 1024 * 1024) -> str:
    md5 = hashlib.md5()
    with tarball_path.open("rb") as fh:
        data = fh.read(buffer_size)
        while data:
            md5.update(data)
            data = fh.read(buffer_size)
    return md5.hexdigest()


def untar(tarball_path: Path, output_path: Path):
    try:
        with tarball_path.open("rb") as fh_in, tarfile.open(
            fileobj=fh_in, mode="r:gz"
        ) as tar_file:
            tar_file.extractall(path=str(output_path))

        tarpath = os.path.join(output_path, VERSION_DICTIONARY[VERSION]["dir_name"])

        # Get a list of all files in the source directory
        files_to_move = [
            f for f in os.listdir(tarpath) if os.path.isfile(os.path.join(tarpath, f))
        ]

        # Move each file to the destination directory
        for file_name in files_to_move:
            source_path = os.path.join(tarpath, file_name)
            destination_path = os.path.join(output_path, file_name)
            shutil.move(source_path, destination_path)
        # remove the directort
        remove_directory(tarpath)

    except OSError:
        logger.error(f"Could not extract {tarball_path} to {output_path}")


def instantiate_dir(db_dir):
    if os.path.isdir(db_dir) == False:
        os.mkdir(db_dir)


def check_db_installation(db_dir):
    downloaded_flag = True
    # PHROGS files
    for file_name in PHROG_DB_NAMES:
        path = os.path.join(db_dir, file_name)
        if os.path.isfile(path) == False:
            logger.info(f"PHROGs Database file {file_name} is missing.")
            downloaded_flag = False
            break
    # VFDB
    for file_name in VFDB_DB_NAMES:
        path = os.path.join(db_dir, file_name)
        if os.path.isfile(path) == False:
            logger.info("VFDB Databases are missing.")
            downloaded_flag = False
            break
    # CARD
    for file_name in CARD_DB_NAMES:
        path = os.path.join(db_dir, file_name)
        if os.path.isfile(path) == False:
            logger.info("CARD Databases are missing.")
            downloaded_flag = False
            break
    # annot.tsv
    path = os.path.join(db_dir, "phrog_annot_v4.tsv")
    if os.path.isfile(path) == False:
        logger.info("PHROGs Annotation File is missing.")
        downloaded_flag = False

    # mash files
    path = os.path.join(db_dir, VERSION_DICTIONARY[VERSION]["inphared_annot"])
    if os.path.isfile(path) == False:
        logger.info("INPHARED Mash Annotation File is missing.")
        downloaded_flag = False

    # mash files
    path = os.path.join(db_dir, VERSION_DICTIONARY[VERSION]["inphared_mash"])
    if os.path.isfile(path) == False:
        logger.info("INPHARED Mash Sketch File is missing.")
        downloaded_flag = False

    return downloaded_flag
